import argparse
import logging
from tado.offsets import TadoOffsetsManager

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-d",
        "--dry-run",
        dest="dry_run",
        action="store_true",
        help="If set, will not make any permanent changes",
    )
    parser.add_argument(
        "-f",
        "--offsets-file",
        dest="offsets_file",
        action="store",
        help="YAML file containing target offsets to set",
    )

    args = parser.parse_args()

    logging.basicConfig(
        level="INFO", format="%(asctime)s [%(name)s] %(levelname)s - %(message)s"
    )

    om = TadoOffsetsManager(offsets_file=args.offsets_file)
    om.apply_offset_changes(dry_run=args.dry_run)



import os
import requests
import yaml


def get_environment_variable(key: str) -> str:
    """Retrieves an environment variable by name.

    Args:
      ``key`` (str):
         Environment variable name

    Returns:
      str: Environment variable value

    Raises:
      ``KeyError``: If the environment variable cannot be found

    """
    try:
        env_var = os.environ[key]
    except KeyError:
        raise KeyError(f"Unable to retrieve environment: {key}")
    else:
        return env_var


def has_token_expired(response: requests.Response) -> bool:
    """Evaluates an API response to determine if an expired token was the cause of
    an *HTTP 401 Unauthorized* error.

    Args:
      ``response`` (requests.Response):
         Response: https://requests.readthedocs.io/en/latest/api/#requests.Response

    Returns:
      bool: ``True`` if an expired token was used, ``False`` if not

    """
    if response.status_code == 401:
        try:
            if any(
                d.get("title", None) == "access token is expired"
                for d in response.json()["errors"]
            ):
                return True
            else:
                return False
        except KeyError:
            return False


def load_yaml_file(filename: str) -> dict:
    """Loads a given YAML file.

    Args:
      ``filename`` (str):
         Full path and filename to load

    Returns:
      dict: Dictionary representation of the file

    """
    with open(filename, "r") as stream:
        return yaml.safe_load(stream)


def mask(value: str, length: int = 4) -> str:
    """Truncates a value, typically for display purposes (e.g. hiding a device's
    serial number from being displayed in full).

    Args:
      ``value`` (str):
         Input string to mask/truncate
      ``length`` (int, optional):
         String length to preserve, defaults to 4

    Returns:
      str: Truncated string suffixed by an ellipsis

    """
    return f"{str(value)[0:length]}..."



from enum import Enum


class EnvVarNames(Enum):
    """Environment variable names used for configuration purposes"""

    CLIENT_ID = "CLIENT_ID"
    CLIENT_SECRET = "CLIENT_SECRET"
    TADO_USERNAME = "TADO_USERNAME"
    TADO_PASSWORD = "TADO_PASSWORD"


class BaseUrls(Enum):
    """Base URL prefixes for various Tado APIs"""

    TADO_AUTH_API = "https://auth.tado.com/oauth/token"
    TADO_BASE_API = "https://my.tado.com/api/v2/me"
    TADO_HOME_API = "https://my.tado.com/api/v2/homes"
    TADO_DEVICE_API = "https://my.tado.com/api/v2/devices"


class AuthenticationProperties(Enum):
    """Static properties used during authentication with Tado APIs"""

    GRANT_TYPE = "password"
    SCOPE = "home.user"



import pytest
import responses
import sys

sys.path.append("../")

import tado.base  # noqa
from tado.enums import AuthenticationProperties, BaseUrls  # noqa


@pytest.fixture
def tado_manager():
    return tado.base.TadoManager(
        load_config=False,
        client_secret="dummy_secret",
        password="dummy_password",
        username="dummy_user",
    )


@pytest.fixture
def mock_responses():
    # Mocked call to get ACCESS_TOKEN
    responses.add(
        responses.POST,
        BaseUrls.TADO_AUTH_API.value,
        json={"access_token": "returned_access_token"},
        status=200,
    )
    # Mocked call to get HOME_ID
    responses.add(
        responses.GET,
        BaseUrls.TADO_BASE_API.value,
        json={"homes": [{"id": 999999, "name": "Home"}]},
        status=200,
    )
    # Mocked call to get LEADER DEVICES
    responses.add(
        responses.GET,
        f"{BaseUrls.TADO_HOME_API.value}/999999/zones",
        json=[
            {
                "id": 1,
                "name": "Lounge",
                "devices": [
                    {
                        "serialNo": "111111111111",
                        "duties": ["ZONE_UI", "ZONE_DRIVER", "ZONE_LEADER"],
                    },
                    {"serialNo": "222222222222", "duties": ["ZONE_UI"]},
                ],
            },
            {
                "id": 3,
                "name": "Kitchen",
                "devices": [
                    {
                        "serialNo": "333333333333",
                        "duties": ["ZONE_UI", "ZONE_DRIVER", "ZONE_LEADER"],
                    }
                ],
            },
        ],
        status=200,
    )


class TestAPIRequests(object):
    @responses.activate
    def test_get_access_token(self, tado_manager, mock_responses):
        tado_manager.get_access_token()
        assert tado_manager.access_token == "returned_access_token"

    @responses.activate
    def test_get_home_id(self, tado_manager, mock_responses):
        tado_manager.get_home_id()
        assert tado_manager.home_id == 999999

    @responses.activate
    def test_get_leader_devices(self, tado_manager, mock_responses):
        tado_manager.home_id = 999999
        tado_manager.get_leader_devices()
        assert tado_manager.leader_devices == {
            "Lounge": "111111111111",
            "Kitchen": "333333333333",
        }


class TestConfiguration(object):
    def test_build_authentication_params(self, tado_manager):
        params = tado_manager.build_authentication_params()
        param_dict = dict(
            param_str_pair.split("=") for param_str_pair in params.split("&")
        )
        assert param_dict["client_id"] == "tado-web-app"
        assert param_dict["client_secret"] == "dummy_secret"
        assert param_dict["grant_type"] == AuthenticationProperties.GRANT_TYPE.value
        assert param_dict["password"] == "dummy_password"
        assert param_dict["scope"] == AuthenticationProperties.SCOPE.value
        assert param_dict["username"] == "dummy_user"

    def test__prepare_headers(self, tado_manager):
        # Headers prior to access token
        headers = tado_manager._prepare_headers()
        assert len(headers) == 1
        assert "Content-Type" in headers.keys()
        assert "Authorization" not in headers.keys()

        # Headers after access token
        tado_manager.access_token = "access_token"
        headers = tado_manager._prepare_headers()
        assert len(headers) == 2
        assert "Content-Type" in headers.keys()
        assert "Authorization" in headers.keys()
        assert headers["Authorization"] == "Bearer access_token"


class TestStatics(object):
    @pytest.mark.parametrize(
        "input, output",
        [(0, 32), (0, 32.0), (0.0, 32), (0.0, 32.0), (32.7, 90.9), (-100, -148)],
    )
    def test_celsius_to_fahrenheit(self, input, output, tado_manager):
        assert tado_manager.celsius_to_fahrenheit(input) == output

    @pytest.mark.parametrize(
        "input, output",
        [(32, 0), (32.0, 0), (32, 0.0), (32.0, 0.0), (90.9, 32.7), (-148, -100)],
    )
    def test_fahrenheit_to_celsius(self, input, output, tado_manager):
        assert tado_manager.fahrenheit_to_celsius(input) == output

    @pytest.mark.parametrize(
        "input, exception_type",
        [(None, TypeError), ("aaa", ValueError), (b"0xFF", ValueError)],
    )
    def test_celsius_to_fahrenheit_invalid_input_type(
        self, input, exception_type, tado_manager
    ):
        with pytest.raises(exception_type) as exc_info:  # noqa
            _ = tado_manager.celsius_to_fahrenheit(input)

    @pytest.mark.parametrize(
        "input, exception_type",
        [(None, TypeError), ("aaa", ValueError), (b"0xFF", ValueError)],
    )
    def test_fahrenheit_to_celsius_invalid_input_type(
        self, input, exception_type, tado_manager
    ):
        with pytest.raises(exception_type) as exc_info:  # noqa
            _ = tado_manager.fahrenheit_to_celsius(input)



import json
import jsonpath_rw_ext as jp
import logging
import pyfiglet
import requests
import sys
import urllib.parse
from retry import retry

from tado.enums import AuthenticationProperties, BaseUrls, EnvVarNames
from tado import utils

log = logging.getLogger(__name__)


class TokenExpired(Exception):
    """Exception raised when an API request is made with an expired token"""

    pass


class TadoManager:
    """Base class used to orchestrate interaction with the Tado API.

    Authentication with the Tado API uses credentials in the following order:

    1. Via **arguments** supplied to this class during instantiation

       * ``username``, ``password``, ``client_id``, ``client_secret``

    2. Via **environment variables** configured by the user

       * ``TADO_USERNAME``, ``TADO_PASSWORD``, ``CLIENT_ID``, ``CLIENT_SECRET``

    Args:
      ``username`` (str, optional):
         Tado account username, defaults to None
      ``password`` (str, optional):
         Tado account password, defaults to None
      ``client_secret`` (str, optional):
         Tado OAuth client secret, defaults to None
      ``client_id`` (str, optional):
         Tado OAuth client identifier, defaults to "tad-web-app"
      ``load_config`` (bool, optional):
         Automatically authenticate and load initial config, defaults to ``True``
    """

    def __init__(
        self,
        username: str = None,
        password: str = None,
        client_secret: str = None,
        client_id: str = "tado-web-app",
        load_config: bool = True,
    ) -> None:
        """Constructor method"""
        self.username = username
        self.password = password
        self.client_secret = client_secret
        self.client_id = client_id
        self.access_token = None
        self.home_id = None
        self.zones = None
        self.leader_devices = None
        if load_config:
            self._load_config()

    def _load_config(self) -> None:
        """Bootstrap configuration method.

        This runs the minimum set of calls to render the package functional (i.e.
        authentication with the Tado API, and retrieval of basic configuration
        information).
        """
        self._show_banner()
        self.get_access_token()
        self.get_home_id()
        self.get_zones()
        self.get_leader_devices()

    def _show_banner(self) -> None:
        """Output ASCII art banner"""
        print(pyfiglet.figlet_format("tado-overseer"))

    def _prepare_headers(self) -> dict:
        """Prepares the set of web request headers for each API call.

        If the ``access_token`` has already been retrieved, it will be included
        in the headers as a bearer authentication token.

        Returns:
          dict: Dict of headers

        """
        headers = {}
        headers["Content-Type"] = "application/json"
        if self.access_token:
            headers["Authorization"] = f"Bearer {self.access_token}"
        return headers

    @retry(TokenExpired, tries=2)
    def _call_tado_api(self, method: str, url: str, data: dict = {}) -> str:
        """Convenience wrapper method to call API endpoints with automatic retry.

        Args:
          ``method`` (str):
             HTTP method (e.g. ``GET``, ``POST``, ``PUT``)
          ``url`` (str):
             URL as constructed and requested by the method caller
          ``data`` (dict, optional):
             Request body, defaults to ``{}``

        Returns:
           str: Stringified JSON response from the called API

        Raises:
          ``TokenExpired``: Raised when an API request uses an expired token

        """
        try:
            response = requests.request(
                method, url, headers=self._prepare_headers(), data=json.dumps(data)
            )
        except Exception:
            log.exception("Exception occurred while calling API")
        else:
            if utils.has_token_expired(response):
                self.get_access_token()
                raise TokenExpired("Token expired, retrying")
            elif response.status_code != 200:
                log.error(
                    f"API request unsuccessful [HTTP code: {response.status_code}]"
                )
                sys.exit(-1)
            else:
                resp_json = response.json()
                return resp_json

    def build_authentication_params(self) -> str:
        """Prepares the set of query parameters required to retrieve an access token
        from the Tado API.

        * **Client ID** - e.g. ``tado-web-app``
        * **Client Secret** - as retrieved from ``https://app.tado.com/env.js``
        * **Grant Type** - e.g. ``password``
        * **Scope** - e.g. ``home.user``
        * **Tado username** - sourced from class instantiation argument or environment
        * **Tado password** - sourced from class instantiation argument or environment

        Returns:
           str: URL encoded string of query parameters
        """
        auth_environment = dict(
            grant_type=AuthenticationProperties.GRANT_TYPE.value,
            scope=AuthenticationProperties.SCOPE.value,
            client_id=self.client_id
            if self.client_id
            else utils.get_environment_variable(EnvVarNames.CLIENT_ID.value),
            client_secret=self.client_secret
            if self.client_secret
            else utils.get_environment_variable(EnvVarNames.CLIENT_SECRET.value),
            password=self.password
            if self.password
            else utils.get_environment_variable(EnvVarNames.TADO_PASSWORD.value),
            username=self.username
            if self.username
            else utils.get_environment_variable(EnvVarNames.TADO_USERNAME.value),
        )
        return urllib.parse.urlencode(auth_environment)

    def get_access_token(self) -> None:
        """Generates and uses a new authorization token for all subsequent requests."""
        log.info("Retrieving ACCESS TOKEN")
        url = "?".join(
            [BaseUrls.TADO_AUTH_API.value, self.build_authentication_params()]
        )
        try:
            access_token = self._call_tado_api("POST", url)["access_token"]
        except KeyError:
            log.exception("Unable to extract ACCESS TOKEN from API response")
        else:
            log.info(f"Retrieved ACCESS TOKEN = [{utils.mask(access_token)}]")
            self.access_token = access_token

    def get_home_id(self) -> None:
        """Retrieves the Tado home identifier."""
        log.info("Retrieving HOME ID")
        url = BaseUrls.TADO_BASE_API.value
        try:
            home_id = self._call_tado_api("GET", url)["homes"][0]["id"]
        except KeyError:
            log.exception("Unable to extract HOME ID from API response")
        else:
            log.info(f"Retrieved HOME ID = [{utils.mask(home_id)}]")
            self.home_id = home_id

    def get_zones(self) -> None:
        """Retrieves the list of zones and full configuration for each."""
        log.info("Retrieving ZONES and DEVICES")
        url = f"{BaseUrls.TADO_HOME_API.value}/{self.home_id}/zones"
        zones = self._call_tado_api("GET", url)
        device_count = len(jp.match("$.[*].devices[*].serialNo", zones))
        log.info(f"Retrieved [{len(zones)}] ZONES and [{device_count}] DEVICES")
        self.zones = zones
        return

    def get_leader_devices(self) -> None:
        """Retrieves the set of ``zone_name:device_serial_number`` for the **leader**
        device in each zone (i.e. the device designated to provide the temperature
        reading for its zone).
        """
        log.info("Retrieving LEADER DEVICES")
        url = f"{BaseUrls.TADO_HOME_API.value}/{self.home_id}/zones"
        devices = {
            i["name"]: jp.match("$.devices[?(duties[*]~'ZONE_LEADER')]", i)[0][
                "serialNo"
            ]
            for i in self._call_tado_api("GET", url)
        }
        log.info(
            f"Retrieved LEADER DEVICES = {[utils.mask(i) for i in devices.values()]}"
        )
        self.leader_devices = devices

    @staticmethod
    def celsius_to_fahrenheit(input_temperature: float) -> float:
        """Converts a given temperature from celsius to fahrenheit.

        Args:
          ``input_temperature`` (float):
             Temperature in celsius

        Returns:
           float: Temperature in fahrenheit

        """
        output_temperature = round((float(input_temperature) * float(9 / 5)) + 32, 1)
        return output_temperature

    @staticmethod
    def fahrenheit_to_celsius(input_temperature: float) -> float:
        """Converts a given temperature from fahrenheit to celsius.

        Args:
          ``input_temperature`` (float):
             Temperature in fahrenheit

        Returns:
           float: Temperature in celsius

        """
        output_temperature = round((float(input_temperature) - 32.0) * float(5 / 9), 1)
        return output_temperature



import dictdiffer
import logging
import sys
from pathlib import Path
from tado import utils
from tado.base import TadoManager
from tado.enums import BaseUrls

log = logging.getLogger(__name__)


class TadoOffsetsManager(TadoManager):
    """Manage temperature offsets for zones and devices.

    Each zone in a Tado home has a designated *leader* device which is responsible for
    providing temperature readings for that zone.  Depending on the proximity of the
    device to heat or cold sources (e.g. radiators, windows etc.) it may be necessary
    to "offset" the readings from that device by a certain temperature, to provide a
    more representative/accurate reading for that zone.  This class can be used to
    inspect and set these offsets.

    **Offsets** are defined as pairs of: ``zone_name:temperature_offset_in_celsius``

    Args:
      ``offsets_dict`` (dict: dict, optional):
         Inline dict of offsets to set, defaults to ``{}``

      ``offsets_file`` (str, optional):
         YAML file with offsets to set, defaults to ``config.yaml``

    """

    def __init__(
        self, offsets_dict: dict = {}, offsets_file: str = "config.yaml", **kwargs
    ) -> None:
        """Constructor method"""
        super().__init__(**kwargs)
        self.offsets_file = offsets_file
        self.offsets_dict = offsets_dict
        if self.offsets_dict:
            self.user_config = self.offsets_dict
        elif Path(self.offsets_file).is_file():
            self.user_config = utils.load_yaml_file(self.offsets_file)
        else:
            log.error("No target offsets configured - exiting")
            sys.exit(-1)

    def get_current_device_offsets(self) -> dict:
        """Retrieves the *currently* configured temperature offset values from the
        leader device in each zone.

        Returns:
          dict: Set of ``zone_name:current_offset_temperature`` entries

        """
        room_offsets = {}
        for room_name, serial_no in self.leader_devices.items():
            url = f"{BaseUrls.TADO_DEVICE_API.value}/{serial_no}/temperatureOffset"
            device_offset = self._call_tado_api("GET", url)["celsius"]
            room_offsets[room_name] = device_offset
        return room_offsets

    def get_target_device_offsets(self) -> dict:
        """Retrieves the *target* temperature offset values to be set in each zone,
        as defined by the user in the input configuration file.

        Returns:
          dict: Set of ``zone_name:target_offset_temperature`` entries

        """
        return self.user_config["tado"]["offsets"]

    def apply_offset_changes(self, dry_run=False) -> None:
        """Sets the target offset to any zone which has a different value already set.

        This function compares the set of user-provided (target) offsets with the
        existing (current) offsets for each zone.  For any zone where the target offset
        is different, a ``PUT`` request is issued to the Tado API to change the offset.

        Args:
          ``dry_run`` (bool, optional, optional):
             Doesn't modify offsets when set to ``True``, defaults to ``False``
        """
        for delta in list(
            dictdiffer.diff(
                self.get_target_device_offsets(), self.get_current_device_offsets()
            )
        ):
            if delta[0] == "change":
                # Extract change details
                room_name, target_temp, current_temp = (
                    delta[1],
                    delta[2][0],
                    delta[2][1],
                )

                # Apply correction
                serial_no = self.leader_devices[room_name]
                url = f"{BaseUrls.TADO_DEVICE_API.value}/{serial_no}/temperatureOffset"
                data_payload = dict(celsius=f"{target_temp:.1f}")
                log.info(
                    "Applying change to [{}] - [current:{}, target: {}]".format(
                        room_name.ljust(15), current_temp, target_temp
                    )
                )
                if not dry_run:
                    self._call_tado_api("PUT", url, data_payload)
